﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Data;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace AGifCreator
{
    /// <summary>
    /// Simple wrapper classes to read and write parts of an Animate Gif.
    /// 
    ///     TODO - Quantize image colors into a single optimize color palette
    ///          - Determine if delta frame save space
    ///          
    ///  Main methods - see ImageInfo() and Save()
    ///          
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    class AGif
    {
        // Some constants used to process data stream
        const Byte EndOfImageId = 59;
        const Byte ControlBlockId = 0x21;
        const Byte ImageBlockId = 0x2c;
        const int HeaderSize = 13;
        const int ColorSize = 3;        // byte red, green, blue
        const int ControlSize = 8;

        /*
          The GIF Local Color Table.
          The size of a color table may be any power of two from 2 to 256.
         */
        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct GifColorTable
        {
            public Byte Red;         // Red Global Color Table Data 
            public Byte Green;       // Green Global Color Table Data 
            public Byte Blue;        // Blue Global Color Table Data 

            public static GifColorTable[] Read(BinaryReader br, int nColors)
            {
                byte[] buff = new byte[Marshal.SizeOf(typeof(GifColorTable))* nColors];

                //Loop until we've read enough bytes (usually once) 
                int amt = 0;
                while (amt < buff.Length)
                    amt += br.Read(buff, amt, buff.Length - amt); //Read bytes 

                //Make sure that the Garbage Collector doesn't move our buffer 
                GCHandle handle = GCHandle.Alloc(buff, GCHandleType.Pinned);

                //Marshal the bytes
                GifColorTable[] colors = new GifColorTable[nColors];
                IntPtr ptr = handle.AddrOfPinnedObject();
                int szCtable = Marshal.SizeOf(typeof(GifColorTable));

                for (int cIdx = 0; cIdx < nColors; cIdx++)
                {
                    colors[cIdx] =
                        (GifColorTable)Marshal.PtrToStructure(ptr,
                        typeof(GifColorTable));
                    ptr = new IntPtr(ptr.ToInt32() + szCtable);
                }

                //Give control of the buffer back to the GC 
                handle.Free();

                return colors;
            }
        };

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct GifHeader                        
        {
            public Byte Signature0;         //   00     ID Signature 'G'          
            public Byte Signature1;         //   01     ID Signature 'I'          
            public Byte Signature2;         //   02     ID Signature 'F'          
            public Byte Version0;           //   03     Version Number  '8'       
            public Byte Version1;           //   04     Version Number  '7' or '9'       
            public Byte Version2;           //   05     Version Number  'a'      
            public UInt16 ScreenWidth;      //   06     Logical Screen Width   
            public UInt16 ScreenHeight;     //   08     Logical Screen Height   
            public Byte PackedField;        //   10     Color Information [CrrrSzzz] 
                                            // [01234567]
                                            //  CrrrSzzz
                                            //  bit 0:      Has Color Table
                                            //  bit 1..3    Color Resolution
                                            //  bit 4:      Sort Flag to Color Table
                                            //  bit 5..7    Size of Color Table: 2^(1+n)

            public Byte ColorIndex;         //   11     Background Color Index 
            public Byte AspectRatio;        //   12     Pixel Aspect Ratio     
            // GifColorTable GlobalCT;      //   13...  Color Table (#elements 2..256)  

            public bool HasColorTable
            {
                get { return (this.PackedField & 0x80) != 0; }
            }
            public int ColorCount
            {
                get { return HasColorTable ? 2 << (this.PackedField & 0x07) : 0; }
            }

            public string Signature
            {
                get { return "" + (char)Signature0 + (char)Signature1 + (char)Signature2; }
            }

            public string Version
            {
                get { return "" + (char)Version0 + (char)Version1 + (char)Version2; }
            }

            public bool FromByte(byte[] data)
            {
                //Make sure that the Garbage Collector doesn't move our buffer 
                GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);

                //Marshal the bytes
                this =
                  (GifHeader)Marshal.PtrToStructure(handle.AddrOfPinnedObject(),
                  typeof(GifHeader));

                //Give control of the buffer back to the GC 
                handle.Free();

                return (Signature == "GIF" && Version == "87a" || Version == "89a");
            }


            public bool Read(BinaryReader br)
            {
                byte[] buff = new byte[Marshal.SizeOf(typeof(GifHeader))];

                //Loop until we've read enough bytes (usually once) 
                int amt = 0;
                while (amt < buff.Length)
                    amt += br.Read(buff, amt, buff.Length - amt); //Read bytes 

                return FromByte(buff); 
            }

            public byte[] ToByte()
            {
                Byte[] data = new Byte[Marshal.SizeOf(typeof(GifHeader))];
                GCHandle handle = GCHandle.Alloc(this, GCHandleType.Pinned);
                Marshal.Copy(handle.AddrOfPinnedObject(), data, 0, data.Length);
                handle.Free();
                return data;
            }

            public void Write(BinaryWriter bw)
            {
                Byte[] data = ToByte();
                bw.Write(data, 0, data.Length);
            }
        };

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct GifImageDescriptor
        {                             // 10 bytes (not counting ColorTable)
            Byte ImageSeparator;      // Image Descriptor identifier           
            UInt16 ImageLeft;         // X position of image on the display    
            UInt16 ImageTop;          // Y position of image on the display    
            UInt16 ImageWidth;        // Width of the image in pixels          
            UInt16 ImageHeight;       // Height of the image in pixels         
            Byte PackedField;         // Image and Color Table Data Information
            // GifColorTable LocalCT;    // Local Color Table 2..256                    
        };

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct GifGraphicControlExtension
        {                             // 8 bytes
            public Byte Introducer;          // Extension Introducer (always 21h)    
            public Byte Label;               // Extension Label (always F9h)         
            public Byte BlockSize;           // Size of Extension Block (always 04h) 
            public Byte PackedField;         // Graphic Data Information   [rrrDDDuT] 
                                             //   bit 0..2: Reserved
                                             //   bit 3..5: Disposal Method
                                             //     0=Not Specified, 1=don't dispose, 
                                             //     2=overwrite Background, 4=overwrite prev
                                             //   bit 6:    User Input Flag  (on=wait for user click before)
                                             //   bit 7:    Transparent Color Flag
            public UInt16 DelayTime;         // Delay time in hundredths of a second 
            public Byte ColorIndex;          // Transparent Color Index              
            public Byte Terminator;          // Block Terminator (always 0)  

            public static Byte LabelId() { return 0xf9; }
            public const Byte FlagTransparentBit = 1;
            public const Byte FlagUserInput = 2;
            public const Byte FlagDontDispose = 4;
            public const Byte FlagOwBackground = 8;
            public const Byte FlagOwPrevious = 16;

            public void Init()
            {
                this.Introducer  = ControlBlockId;
                this.Label       = 0xf9;     //249
                this.BlockSize   = 4;
                this.PackedField = FlagTransparentBit; //  +FlagOwPrevious;
                this.DelayTime   = 800;
                this.ColorIndex  = 255;
                this.Terminator  = 0;
            }

            public bool FromByte(byte[] data, int startOffset)
            {
                //Make sure that the Garbage Collector doesn't move our buffer 
                GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);
                IntPtr ptr = new IntPtr(handle.AddrOfPinnedObject().ToInt32() + startOffset);

                //Marshal the bytes
                this =
                  (GifGraphicControlExtension)Marshal.PtrToStructure(ptr,
                  typeof(GifGraphicControlExtension));

                //Give control of the buffer back to the GC 
                handle.Free();

                return (this.Introducer == ControlBlockId && this.Label == 0xf9 && this.BlockSize == 4 &&
                   this.Terminator == 0);
            }


            public bool Read(BinaryReader br)
            {
                byte[] buff = new byte[Marshal.SizeOf(typeof(GifGraphicControlExtension))];

                //Loop until we've read enough bytes (usually once) 
                int amt = 0;
                while (amt < buff.Length)
                    amt += br.Read(buff, amt, buff.Length - amt); //Read bytes 

                return FromByte(buff, 0);
            }

            public byte[] ToByte()
            {
                Byte[] data = new Byte[Marshal.SizeOf(typeof(GifGraphicControlExtension))];
                GCHandle handle = GCHandle.Alloc(this, GCHandleType.Pinned);
                Marshal.Copy(handle.AddrOfPinnedObject(), data, 0, data.Length);
                handle.Free();
                return data;
            }

            public void Write(BinaryWriter bw)
            {
                Byte[] data = ToByte();
                bw.Write(data, 0, data.Length);
            }
        };

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct GifApplicationExtension
        {                            // 19 bytes
            public Byte Introducer;         // Extension Introducer (always 21h)    
            public Byte Label;              // Extension Label (always FFh)         
            public Byte BlockSize;          // Size of Extension Block (always 0Bh) 
            public Byte Identifier0;        // Application Identifier               
            public Byte Identifier1;        // Application Identifier               
            public Byte Identifier2;        // Application Identifier               
            public Byte Identifier3;        // Application Identifier               
            public Byte Identifier4;        // Application Identifier               
            public Byte Identifier5;        // Application Identifier               
            public Byte Identifier6;        // Application Identifier               
            public Byte Identifier7;        // Application Identifier               
            public Byte AuthentCode0;       // Application Authentication Code      
            public Byte AuthentCode1;       // Application Authentication Code      
            public Byte AuthentCode2;       // Application Authentication Code      
            public UInt32 ApplicationData;  // Application data sub-blocks          
            public Byte Terminator;         // Block Terminator (always 0) 

            static Byte LabelId() { return 0xff; }
            public string Identifier
            {
                get { return "" + (char)Identifier0 + (char)Identifier1 + (char)Identifier2 + (char)Identifier3 + (char)Identifier4 + (char)Identifier5 + (char)Identifier6 + (char)Identifier7; }
                set
                {
                    Identifier0 = (Byte)value[0];
                    Identifier1 = (Byte)value[1];
                    Identifier2 = (Byte)value[2];
                    Identifier3 = (Byte)value[3];
                    Identifier4 = (Byte)value[4];
                    Identifier5 = (Byte)value[5];
                    Identifier6 = (Byte)value[6];
                    Identifier7 = (Byte)value[7];
                }
            }

            public string Authentication
            {
                get { return "" + (char)AuthentCode0 + (char)AuthentCode1 + (char)AuthentCode2; }
                set
                {
                    AuthentCode0 = (Byte)value[0];
                    AuthentCode1 = (Byte)value[1];
                    AuthentCode2 = (Byte)value[2];
                }
            }

            public void Init()
            {
                this.Introducer = ControlBlockId;
                this.Label      = 0xff;
                this.BlockSize  = 0xb;
                Identifier = "NETSCAPE";
                Authentication = "2.0";
                this.ApplicationData = 259;
                this.Terminator = 0;
            }

            public bool FromByte(byte[] data)
            {
                //Make sure that the Garbage Collector doesn't move our buffer 
                GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);

                //Marshal the bytes
                this =
                  (GifApplicationExtension)Marshal.PtrToStructure(handle.AddrOfPinnedObject(),
                  typeof(GifApplicationExtension));

                //Give control of the buffer back to the GC 
                handle.Free();

                return (this.Introducer == ControlBlockId && this.Label == 0xff && this.BlockSize == 11 &&
                    this.Terminator == 0);
            }

            public bool Read(BinaryReader br)
            {
                byte[] buff = new byte[Marshal.SizeOf(typeof(GifApplicationExtension))];

                //Loop until we've read enough bytes (usually once) 
                int amt = 0;
                while (amt < buff.Length)
                    amt += br.Read(buff, amt, buff.Length - amt); //Read bytes 

                return FromByte(buff);
            }

            public byte[] ToByte()
            {
                Byte[] data = new Byte[Marshal.SizeOf(typeof(GifApplicationExtension))];
                GCHandle handle = GCHandle.Alloc(this, GCHandleType.Pinned);
                Marshal.Copy(handle.AddrOfPinnedObject(), data, 0, data.Length);
                handle.Free();
                return data;
            }

            public void Write(BinaryWriter bw)
            {
                Byte[] data = ToByte();
                bw.Write(data, 0, data.Length);
            }
        };

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct GifCommentExtension
        {
            Byte   Introducer;         // Extension Introducer (always 21h)    
            Byte   Label;              // Comment Label (always FEh)           
            UInt32 CommentData;        // Comment data sub-blocks              
            Byte   Terminator;         // Block Terminator (always 0) 

            static Byte LabelId() { return 0xfe; }
            public void Init()
            {
                this.Introducer = ControlBlockId;
                this.Label      = 0xfe;
                this.Terminator = 0;
            }
        };


        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct GifImageBlockHeader
        {
            public Byte Introducer;         // always 0x2c
            public UInt16 LeftPos;
            public UInt16 TopPos;
            public UInt16 Width;
            public UInt16 Height;
            public Byte Terminator;

            public bool FromByte(byte[] data)
            {
                //Make sure that the Garbage Collector doesn't move our buffer 
                GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);

                //Marshal the bytes
                this =
                  (GifImageBlockHeader)Marshal.PtrToStructure(handle.AddrOfPinnedObject(),
                  typeof(GifImageBlockHeader));

                //Give control of the buffer back to the GC 
                handle.Free();

                return (this.Introducer == ImageBlockId /* && this.Terminator == 0 */);
            }

            public bool Read(BinaryReader br)
            {
                byte[] buff = new byte[Marshal.SizeOf(typeof(GifImageBlockHeader))];

                //Loop until we've read enough bytes (usually once) 
                int amt = 0;
                while (amt < buff.Length)
                    amt += br.Read(buff, amt, buff.Length - amt); //Read bytes 

                return FromByte(buff);
            }

            public byte[] ToByte()
            {
                Byte[] data = new Byte[Marshal.SizeOf(typeof(GifImageBlockHeader))];
                GCHandle handle = GCHandle.Alloc(this, GCHandleType.Pinned);
                Marshal.Copy(handle.AddrOfPinnedObject(), data, 0, data.Length);
                handle.Free();
                return data;
            }

            public void Write(BinaryWriter bw)
            {
                Byte[] data = ToByte();
                bw.Write(data, 0, data.Length);
            }
        };

        struct GifImageBlock
        {
            public GifImageBlockHeader header;
            public Byte lzwMinCodeSize;
            public Byte[] data;     // n blocks [byte BlockSize followed by bytes] ... 0x00

            public bool Read(BinaryReader br)
            {
                if (header.Read(br))
                {
                    Byte [] buff = new Byte[256];
                    br.Read(buff, 0, 1);
                    lzwMinCodeSize = buff[0];
                    data = new Byte[0];

                    int len = 1;
                    while (len != 0)
                    {
                        br.Read(buff, 0, 1);
                        len = buff[0];

                        if (len != 0)
                        {
                            if (br.Read(buff, 0, len) == len)
                            {
                                Byte[] bigBuff = new Byte[data.Length + len];
                                System.Buffer.BlockCopy(data, 0, bigBuff, 0, data.Length);
                                System.Buffer.BlockCopy(buff, 0, bigBuff, data.Length, len);
                                data = bigBuff;
                            }
                        }
                    }

                    return true;
                }

                return false;
            }
        };


        public static bool ImageInfo(string srcFilename, TextWriter tw)
        {
            FileStream srcStream = File.Open(srcFilename, FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
            BinaryReader binaryReader = new BinaryReader(srcStream);

            GifHeader gifHeader = new GifHeader();
            GifApplicationExtension gifAppExt = new GifApplicationExtension();
            GifGraphicControlExtension gifCtrl = new GifGraphicControlExtension();
            GifImageBlock gifImageBlock = new GifImageBlock();

            if (!gifHeader.Read(binaryReader))
            {
                tw.WriteLine("Failed to read header");
                return false;
            }

            tw.WriteLine(
                "File:     " + srcFilename +
                "\nSignature:" + gifHeader.Signature +
                ", Version:" + gifHeader.Version +
                ", Width:" + gifHeader.ScreenWidth.ToString() +
                ", Height:" + gifHeader.ScreenHeight.ToString() +
                ", PackBits:" + gifHeader.PackedField.ToString() +
                ", ColorIdx:" + gifHeader.ColorIndex.ToString() +
                ", Aspect:" + gifHeader.AspectRatio.ToString()
                );

            GifColorTable[] colors = GifColorTable.Read(binaryReader, gifHeader.ColorCount);

            tw.WriteLine("Colors   Red Green  Blue");
            for (int cIdx = 0; cIdx < colors.Length; cIdx++)
            {
                tw.WriteLine(cIdx.ToString().PadLeft(6) + 
                    colors[cIdx].Red.ToString().PadLeft(6) +
                    colors[cIdx].Green.ToString().PadLeft(6) +
                    colors[cIdx].Blue.ToString().PadLeft(6));
            }

            if (!gifAppExt.Read(binaryReader))
            {
                tw.WriteLine("Failed to read Application Extension");
                return false;
            }

            tw.WriteLine(
                "Identifier:" + gifAppExt.Identifier +
                ", Authentication:" + gifAppExt.Authentication +
                ", ApplicationData:" + gifAppExt.ApplicationData);

            tw.WriteLine("Frame Bits Delay Color  Size Lzw  Left    Top  Width Height");

            int frameIdx = 0;
            while (binaryReader.PeekChar() == 0x21 && 
                gifCtrl.Read(binaryReader) &&
                gifImageBlock.Read(binaryReader))
            {
                frameIdx++;

                tw.WriteLine(
                    frameIdx.ToString().PadLeft(5) + "  " 
                    + gifCtrl.PackedField.ToString("000") 
                    + gifCtrl.DelayTime.ToString().PadLeft(6) 
                    + gifCtrl.ColorIndex.ToString().PadLeft(5) 
                    + gifImageBlock.data.Length.ToString().PadLeft(6) 
                    + gifImageBlock.lzwMinCodeSize.ToString("00").PadLeft(5) 
                    + gifImageBlock.header.LeftPos.ToString().PadLeft(5) 
                    + gifImageBlock.header.TopPos.ToString().PadLeft(7)
                    + gifImageBlock.header.Width.ToString().PadLeft(7) 
                    + gifImageBlock.header.Height.ToString().PadLeft(7)
                    );
            }

            Byte[]  buff = new Byte[256];
            binaryReader.Read(buff, 0, 1);

            tw.WriteLine("EndImage: " + buff[0].ToString());

            // Dont know why some images end in 101
            return (buff[0] == EndOfImageId || buff[0] == 101);
        }

        /// <summary>
        /// Save sequence of image frames as an Animated GIF
        /// </summary>
        /// <param name="dstFilename"></param>
        /// <param name="srcImageList"></param>
        /// <param name="speedMsec">Speed in msec</param>
        /// <returns>True if all went well</returns>
        public static bool Save(string dstFilename, ImageList srcImageList, ushort speedMsec)
        {
            FileStream dstStream = File.Create(dstFilename);
            BinaryWriter binaryWriter = new BinaryWriter(dstStream);
            GifApplicationExtension gifAppExt = new GifApplicationExtension();
            GifHeader imgHeader = new GifHeader();
            GifGraphicControlExtension gifCtrExt = new GifGraphicControlExtension();

            gifAppExt.Init();
            gifCtrExt.Init();
            gifCtrExt.DelayTime = (ushort)(speedMsec/10);

            int frameIdx = 0;
            foreach (Image image in srcImageList.Images)
            {
                MemoryStream memoryStream = new MemoryStream();
                image.Save(memoryStream, ImageFormat.Gif);
                Byte[] imgBuf = memoryStream.ToArray();
                if (!imgHeader.FromByte(imgBuf))
                    return false;

                if (frameIdx++ == 0)
                {
                    // Write header once
                    imgHeader.Write(binaryWriter);  
                    binaryWriter.Write(imgBuf, AGif.HeaderSize, imgHeader.ColorCount * AGif.ColorSize); 
                    gifAppExt.Write(binaryWriter);
                }

                if (!gifCtrExt.FromByte(imgBuf, AGif.HeaderSize + AGif.ColorSize * imgHeader.ColorCount))
                    return false;
                gifCtrExt.DelayTime = (ushort)(speedMsec / 10);
                gifCtrExt.Write(binaryWriter);
                int hdrSize = AGif.HeaderSize + AGif.ColorSize * imgHeader.ColorCount + AGif.ControlSize;
                binaryWriter.Write(imgBuf, hdrSize, imgBuf.Length - hdrSize - 1); 
            }

            binaryWriter.Write((Byte)0x3B); //Image terminator 
            binaryWriter.Close();

            return true;
        }
    }
}
