﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Drawing.Imaging;

namespace AGifCreator
{
    /// <summary>
    /// Simple Animate Gif creator - construct Agif from several image frames.
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public partial class AGifCreatorApp : Form
    {
        #region Data Members
        private int leftPanelWidth = 0;
        private int topPanelHeight = 0;
        private FileTreeExplorer fileTreeExplorer = new FileTreeExplorer();
        private Microsoft.Win32.RegistryKey regKey;
        private string filename = string.Empty;
        private bool fireTreeEvent = true;
        private ushort speedMsec = 800;

        #endregion

        public AGifCreatorApp()
        {
            InitializeComponent();

            this.regKey =
                Microsoft.Win32.Registry.CurrentUser.CreateSubKey(@"Software\AGifCreator");

            this.Text = ProductNameAndVersion();
        }

        public static string ProductNameAndVersion()
        {
            string appName = Application.ProductName;
            string appVern = Application.ProductVersion;
            return appName + " v" + appVern.Substring(0, 3); //  Get part of versoin string "n.n"
        }

        #region Save and Restore layout to/from registry
        /// <summary>
        /// Save layout to registry on close
        /// </summary>
        private void AGifCreator_FormClosing(object sender, FormClosingEventArgs e)
        {
            // Get latest layout settings.
            if (this.leftPanel.Visible)
                this.leftPanelWidth = this.leftRightSplitter.SplitterDistance;
            if (this.fileListExplorer.Visible)
                this.topPanelHeight = this.verticalSplitter.SplitterDistance;

            // Save layout;
            this.regKey.SetValue("leftWidth",   this.leftPanelWidth);
            this.regKey.SetValue("leftVisible", this.leftPanel.Visible);
            this.regKey.SetValue("topHeight",    this.topPanelHeight);
            this.regKey.SetValue("topVisible",  this.fileListExplorer.Visible);

            // Save other UI layout/settings to Registry.
            this.fileListExplorer.SaveToRegistry(this.regKey.CreateSubKey("FileListExplorer"));
        } 

        /// <summary>
        /// Restore layout on Load
        /// </summary>
        private void AGifCreator_Load(object sender, EventArgs e)
        {
            // Restore layout;
            this.leftPanelWidth = (int)this.regKey.GetValue("leftWidth", this.leftRightSplitter.SplitterDistance);
            this.leftFileTreeToolStripMenuItem.Checked = 
                this.leftPanel.Visible = Convert.ToBoolean(this.regKey.GetValue("leftVisible", "true"));
            this.topPanelHeight = (int)this.regKey.GetValue("topHeight", this.verticalSplitter.SplitterDistance);
            this.topFileListToolStripMenuItem.Checked = 
                this.fileListExplorer.Visible = Convert.ToBoolean(this.regKey.GetValue("topVisible", "true"));

            if (this.leftPanel.Visible == false)
                this.leftRightSplitter.SplitterDistance = 0;
            else
                this.leftRightSplitter.SplitterDistance = this.leftPanelWidth;

            if (this.fileListExplorer.Visible == false)
                this.verticalSplitter.SplitterDistance = 0;
            else
                this.verticalSplitter.SplitterDistance = this.topPanelHeight;

            // Load file tree
            fileTreeExplorer.CreateTree(this.tvFileExplorer);

            // Load other UI layout/settings from Registry.
            this.fileListExplorer.LoadFromRegistry(this.regKey.CreateSubKey("FileListExplorer"));
        }
        #endregion

        #region Button controls
        /// <summary>
        ///  Add files selected in file explorer to image explorer (aka Frame list)
        /// </summary>
        private void AddBtn_Click(object sender, EventArgs e)
        {
            foreach (FileInfo fileInfo in this.fileListExplorer.SelectedItems)
            {
                this.imageListExplorer.Append(fileInfo);
            }
        }

        /// <summary>
        ///  Clear image explorer.
        /// </summary>
        private void newBtn_Click(object sender, EventArgs e)
        {
            this.imageListExplorer.Clear();
            this.filename = string.Empty;
            this.UpdateTitle();
        }

        /// <summary>
        ///  Delete selected images from image explorer
        /// </summary>
        private void delBtn_Click(object sender, EventArgs e)
        {
            this.imageListExplorer.DeleteSelected();
        }

        /// <summary>
        ///  Exit application
        /// </summary>
        private void exitBtn_Click(object sender, EventArgs e)
        {
            this.Close();
        }
        /// <summary>
        /// Update title bar with name of current image
        /// </summary>
        private void UpdateTitle()
        {
            if (this.filename.Length > 0)
                this.Text = "AGifCreator - File:" + this.filename;
            else
                this.Text = "AGifCreator";
        }

        /// <summary>
        /// Display information about image. Current image if any, else prompt for image.
        /// </summary>
        private void imageInfoBtn_Click(object sender, EventArgs e)
        {
            string fn = this.filename;
            if (fn.Length == 0 && this.openFileDialog.ShowDialog() != DialogResult.OK)
                return;

            if (fn.Length == 0)
                fn = this.openFileDialog.FileName;

            TextWriter tw = new StringWriter();
            AGif.ImageInfo(fn, tw);

            InfoDialog infoForm = new InfoDialog();
            infoForm.Text = "Information for:" + fn;
            infoForm.textBox.Text = tw.ToString();

            infoForm.Show();
        }

        /// <summary>
        /// Load image (animated image) into image explorer.
        /// If animated GIF, insert each frame separately.
        /// </summary>
        private void LoadBtn_Click(object sender, EventArgs e)
        {
            if (this.openFileDialog.ShowDialog() == DialogResult.OK)
            {
                this.imageListExplorer.Clear();
                foreach (string fn in this.openFileDialog.FileNames)
                {
                    if (this.imageListExplorer.Append(new FileInfo(fn)))
                    {
                        this.filename = fn;
                        this.UpdateTitle();
                    }
                    else
                        this.statusBox.Text = "Failed to load " + fn;
                }
            }
        }

        /// <summary>
        /// Save current frame list to current image name.
        /// If no 'current' image specified, prompt for a filename.
        /// </summary>
        private void saveBtn_Click(object sender, EventArgs e)
        {
            string fn = this.filename;

            // If no active filename, prompt for new file.
            if (fn.Length == 0 &&
              this.saveFileDialog.ShowDialog() != DialogResult.OK)
                return;

            if (fn.Length == 0)
                fn = this.saveFileDialog.FileName;

            try
            {
                AGif.Save(fn, this.imageListExplorer.ImageList, this.speedMsec);
                this.filename = fn;
                this.UpdateTitle();
                this.statusBox.Text = "Saved to " + fn;
            }
            catch
            {
                this.statusBox.Text = "Failed to save to " + fn;
            }
        }

        /// <summary>
        ///  Save current frame list as an animated gif.
        /// </summary>
        private void SaveAsBtn_Click(object sender, EventArgs e)
        {
            if (this.saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    this.speedMsec = this.framePlayer.speedMsec;
                    AGif.Save(this.saveFileDialog.FileName, this.imageListExplorer.ImageList, this.speedMsec);
                    this.filename = this.saveFileDialog.FileName;
                    this.UpdateTitle();
                }
                catch
                {
                    this.statusBox.Text = "Failed To save to " + this.saveFileDialog.FileName;
                }
            }
        }

        /// <summary>
        /// Capture screen and add to image explorer.
        /// </summary>
        private void captureBtn_Click(object sender, EventArgs e)
        {
            CaptureDialog captureDialog = new CaptureDialog();
            this.Hide();
            captureDialog.ShowDialog();
            this.Show();

            int idx = 0;
            foreach (Image image in captureDialog.imageList)
            {
                FileInfo fileInfo = new FileInfo("screen"+idx++.ToString() +".bmp");
                this.imageListExplorer.Append(fileInfo, image, -1);
            }
        }

        /// <summary>
        /// Display about box.
        /// </summary>
        private void logo_Click(object sender, EventArgs e)
        {
            AboutDialog aboutDialog = new AboutDialog();
            aboutDialog.Show();
        }

        #endregion

        #region View Controls to open close panels
        /// <summary>
        ///  User closing file tree panel.
        /// </summary>
        private void closeFileTreeBtn_Click(object sender, EventArgs e)
        {
            leftFileTreeToolStripMenuItem_Click(sender, e);
        }

        /// <summary>
        /// Open or close the left "tree" panel
        /// </summary>
        private void leftFileTreeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.leftFileTreeToolStripMenuItem.Checked = !this.leftFileTreeToolStripMenuItem.Checked;

            this.leftPanel.Visible = this.leftFileTreeToolStripMenuItem.Checked;

            if (this.leftPanel.Visible)
            {
                this.leftRightSplitter.SplitterDistance = this.leftPanelWidth;
            }
            else
            {
                this.leftPanelWidth = this.leftRightSplitter.SplitterDistance;
                this.leftRightSplitter.SplitterDistance = 0;
            }
        }

        /// <summary>
        ///  Open or close the top file explorer
        /// </summary>
        private void topFileListToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.topFileListToolStripMenuItem.Checked = !this.topFileListToolStripMenuItem.Checked;

            this.fileListExplorer.Visible = this.topFileListToolStripMenuItem.Checked;

            if (this.fileListExplorer.Visible)
            {
                this.verticalSplitter.SplitterDistance = this.topPanelHeight;
            }
            else
            {
                this.topPanelHeight = this.verticalSplitter.SplitterDistance;
                this.verticalSplitter.SplitterDistance = 0;
            }
        }
        #endregion

        #region Manage Tree view navigation
        // Populate tree node on expand.
        private void tvFileExplorer_BeforeExpand(object sender, TreeViewCancelEventArgs e)
        {
            if (e.Node.Nodes[0].Text == "")
            {
                System.Diagnostics.Debug.WriteLine("Expand tree node:" + e.Node.FullPath);
                TreeNode node = fileTreeExplorer.EnumerateDirectory(e.Node);
            }          
        }

        /// <summary>
        ///  Click on tree node adds file or directory to File list explorer
        /// </summary>
        private void tvFileExplorer_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            // remove "Desktop\"
            string fullPath = (e.Node.FullPath.Length > 8) ? e.Node.FullPath.Substring(8) : @"\"; 
            DirectoryInfo nodeDir = new DirectoryInfo(fullPath);
            if (nodeDir.Exists == false)
                return;

            this.fireTreeEvent = false;
            this.fileListExplorer.Add(nodeDir);
            this.fireTreeEvent = true;
        }
        #endregion

        #region Manage File List Explorer navigation
        /// <summary>
        /// File List explorer changed - update tree view to follow along.
        /// </summary>
        private void fileListExplorer_Changed(object sender, EventArgs e)
        {
            DirectoryInfo dirInfo = this.fileListExplorer.Directory;
            if (this.fireTreeEvent)
                this.fileTreeExplorer.ExpandNode(this.tvFileExplorer, dirInfo);
        }

        /// <summary>
        /// Space bar or double click in file explorer causes item to be added to image explorer.
        /// </summary>
        private void fileListExplorer_DoubleClicked(object sender, EventArgs e)
        {
            AddBtn_Click(sender, e);
        }
        #endregion

        /// <summary>
        /// Display image under cursor.
        /// </summary>
        private void imageListExplorer_HoverChanged(object sender, EventArgs e)
        {
            ListViewItemMouseHoverEventArgs eHover = e as ListViewItemMouseHoverEventArgs;
            this.framePlayer.ViewImage = this.imageListExplorer.ImageFor(eHover.Item);
            int frame = Int32.Parse(eHover.Item.Text);
            this.framePlayer.ViewFrameMax = this.imageListExplorer.Count - 1;
            this.framePlayer.ViewFrame = frame;
        }

        /// <summary>
        /// Player control has changed, keep image explorer (frame in view)
        /// </summary>
        private void framePlayer_Changed(object sender, EventArgs e)
        {
            FrameControl.EventFrame evtFrame = e as FrameControl.EventFrame;
            this.imageListExplorer.Frame = evtFrame.Frame;
        }
    }
}
