﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;


namespace AGifCreator
{
    /// <summary>
    /// File directory browsing combobox
    ///     Use TAB to complete partial directory.
    ///     Accepts dropping directories
    ///     Persist combobox items in registry
    /// 
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// </summary>
    public partial class FileCombo : UserControl
    {
        public FileCombo()
        {
            InitializeComponent();

            // Enable auto file system completion.
            this.comboBox.AutoCompleteMode = AutoCompleteMode.SuggestAppend;
            this.comboBox.AutoCompleteSource = AutoCompleteSource.FileSystem;
        }

        public void LoadFromRegistry(Microsoft.Win32.RegistryKey key)
        {
            int count = (int)key.GetValue("FileComboCount", 0);
            for (int idx = 0; idx < count; idx++)
            {
                this.comboBox.Items.Add(key.GetValue("FileCombo" + idx.ToString("00"), ""));
            }

            if (this.comboBox.Items.Count > 0)
                this.comboBox.SelectedItem = this.comboBox.Items[0];
        }

        public void SaveToRegistry(Microsoft.Win32.RegistryKey key)
        {
            // Save 1st 10 items.
            int cnt = Math.Min(10, this.comboBox.Items.Count);
            key.SetValue("FileComboCount", cnt);
            for (int idx = 0; idx < cnt; idx++)
            {
                key.SetValue("FileCombo"+idx.ToString("00"), this.comboBox.Items[idx]);
            }
        }

#if false
        /// <summary>
        /// Emulate combo-box fileSystem auto complete.
        ///   Tab shows next match
        ///   Slash advance to next sub-directory.
        /// </summary>
        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            if (keyData == Keys.Tab && comboBox.Focused)
            {
                // Get name pieces left and right of tab position.
                string leftName = this.comboBox.Text.Substring(0, this.tabPos);
                string rightName = string.Empty;
                if (this.tabPos < this.comboBox.Text.Length)
                    rightName = Path.GetFileName(this.comboBox.Text);

                // Parse parts of current filepath.
                string rootDir = "";
                try
                {
                    if (leftName.Length > 0)
                        rootDir = Path.GetDirectoryName(leftName);
                    if (rootDir == null)
                        rootDir = Path.GetPathRoot(leftName);
                }
                catch { }
                string partial = Path.GetFileName(leftName);

                bool foundRight = (rightName.Length == 0);             

                try
                {
                    DirectoryInfo dir = new DirectoryInfo(rootDir);
                    foreach (FileInfo fileInfo in dir.GetFiles())
                    {
                        if (fileInfo.Name.StartsWith(partial, StringComparison.OrdinalIgnoreCase))
                        {
                            if (!foundRight)
                                foundRight = (rightName == fileInfo.Name);
                            else
                            {
                                this.comboBox.Text = rootDir + fileInfo.Name;
                                this.comboBox.SelectionStart = this.tabPos; 
                                this.insertMode = false;
                                return true;
                            }
                        }
                    }

                    foreach (DirectoryInfo dirInfo in dir.GetDirectories())
                    {
                        if (dirInfo.Name.StartsWith(partial, StringComparison.OrdinalIgnoreCase))
                        {
                            if (!foundRight)
                                foundRight = (rightName == dirInfo.Name);
                            else
                            {
                                this.comboBox.Text = dirInfo.FullName;
                                this.comboBox.SelectionStart = this.tabPos; 
                                this.insertMode = false;
                                return true;
                            }
                        }
                    }
                }
                catch { }

                if (this.tabPos < this.comboBox.Text.Length)
                {
                    this.comboBox.Text = this.comboBox.Text.Remove(this.tabPos);
                    this.comboBox.SelectionStart = this.tabPos;
                }
                return true;
            }
            else
            {
                if (keyData == Keys.Insert)
                    this.insertMode = !this.insertMode;
                return base.ProcessCmdKey(ref msg, keyData);
            }
        }
#endif

        private void comboBox_TextUpdate(object sender, EventArgs e)
        {
            this.tabPos = Math.Min(this.comboBox.SelectionStart, this.comboBox.Text.Length);
        }

        private bool isAlphaNum(KeyEventArgs e)
        {
            Keys k = e.KeyCode;
            bool isAlpha = !e.Alt && !e.Control && (k >= Keys.A && k <=Keys.Z);
            // Determine whether the keystroke is a number
            bool d0d9 = k >= Keys.D0 && k <= Keys.D9;
            bool numpad = k >= Keys.NumPad0 && k <= Keys.NumPad9;
            bool isNumber = (!e.Alt && !e.Control && !e.Shift) && (d0d9 || numpad);

            return (isAlpha || isNumber);
        }

        /// <summary>
        ///  Allow TAB to work as input (and not advance to next UI)
        /// </summary>
        private void comboBox_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            if (e.KeyCode == Keys.Tab)
            {
                e.IsInputKey = true;
            }
        }

        /// <summary>
        /// Manually deal with insert/overtype mode
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void comboBox_KeyDown(object sender, KeyEventArgs e)
        {
            int dirSlash = (int)Path.DirectorySeparatorChar + 128;
            
            if (e.KeyCode == Keys.Enter)
            {
                if (this.Changed != null)
                    this.Changed(this, EventArgs.Empty);
                this.comboBox.Items.Remove(this.comboBox.Text);
                this.comboBox.Items.Insert(0, this.comboBox.Text);
                e.SuppressKeyPress = true;
            }
            else if (isAlphaNum(e) && !this.insertMode && this.comboBox.SelectionStart <
                this.comboBox.Text.Length)
            {
                int start = this.comboBox.SelectionStart;
                this.comboBox.Text = this.comboBox.Text.Remove(this.comboBox.SelectionStart);
                this.comboBox.SelectionStart = start;
            }
            else if (e.KeyValue == dirSlash || e.KeyCode == Keys.Tab)
            {
                if (e.KeyCode == Keys.Tab)
                    e.SuppressKeyPress = true;
               
                try
                {
                    DirectoryInfo dirInfo = new DirectoryInfo(this.comboBox.Text);
                    if (dirInfo.Exists)
                        this.comboBox.SelectionStart = this.comboBox.Text.Length;
                }
                catch { }
            }
        }

        public void comboBox_DragDrop(object sender, DragEventArgs e)
        {
            try
            {
                string fileName = "";
                if (e.Data.GetDataPresent(DataFormats.Text))
                {
                    fileName = (string)e.Data.GetData(DataFormats.Text);
                }
                else if (e.Data.GetDataPresent(DataFormats.FileDrop))
                {
                    string[] fileNames;
                    fileNames = (string[])e.Data.GetData(DataFormats.FileDrop);
                    if (fileNames != null)
                    {
                        fileName = fileNames[0];
                    }
                }

                if (fileName != null && fileName.Length > 0)
                {
                    DirectoryInfo dirInfo = new DirectoryInfo(fileName);
                    if (dirInfo.Exists)
                    {
                        this.comboBox.Items.Remove(dirInfo.FullName);
                        this.comboBox.Items.Insert(0, dirInfo.FullName);
                        this.comboBox.SelectedItem = this.comboBox.Items[0];
                    }
                }
            }
            catch { }
        }

        public void comboBox_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.Text) ||
                e.Data.GetDataPresent(DataFormats.FileDrop))
                e.Effect = DragDropEffects.Copy;
            else
                e.Effect = DragDropEffects.None;
        }

        private void comboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Place new selection at top of item list.
            string newValue = this.comboBox.Text;
            int pos = this.comboBox.FindStringExact(newValue, -1);
            if (pos != 0)
            {
                this.comboBox.Items.RemoveAt(pos);
                this.comboBox.Items.Insert(0, newValue);
                // Following line will cause recursion back into this method
                this.comboBox.SelectedIndex = 0;  
                return;
            }

            if (this.Changed != null)
                this.Changed(this, EventArgs.Empty);
        }

        public string FileName
        {
            get { return this.comboBox.Text; }
            set { this.comboBox.Text = value; }
        }

        private int tabPos = 0;
        private bool insertMode = true;
        public event EventHandler Changed;

        /// <summary>
        ///  Start Drag action.
        /// </summary>
        private void icon_MouseMove(object sender, MouseEventArgs e)
        {
            DataObject dataObject = new DataObject(DataFormats.FileDrop, FileName);
            DoDragDrop(dataObject, DragDropEffects.Copy);
        }

      
    }
}
