﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace AGifCreator
{
    /// <summary>
    /// File list explorer with:
    ///    1. Combo bar which auto completes file directories on TAB
    ///    2. FilePath bar which peels back by clicking on subdirectories
    ///    3. File list which supports standard views
    ///    4. Sortable columns
    ///    5. Drag and drop
    ///    
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public partial class FileListExplorer : UserControl
    {
        public FileListExplorer()
        {
            InitializeComponent();
            ViewBtn_Click(this.detailViewBtn, EventArgs.Empty);

            // Create an instance of a ListView column sorter and assign it 
            // to the ListView control.
            this.columnSorter = new ListViewColumnSorter();
            this.fileView.ListViewItemSorter = this.columnSorter;
        }

        public void LoadFromRegistry(Microsoft.Win32.RegistryKey key)
        {
            this.fileComboBar.LoadFromRegistry(key.CreateSubKey("FileComboBar"));
        }

        public void SaveToRegistry(Microsoft.Win32.RegistryKey key)
        {
            this.fileComboBar.SaveToRegistry(key.CreateSubKey("FileComboBar"));
        }

        #region User Interface callbacks
        public void ViewStyle(bool menuBarOn, bool comboOn, bool pathBarOn)
        {
            this.menuBar.Visible = menuBarOn;
            this.fileComboBar.Visible = comboOn;
            this.filePathBar.Visible = pathBarOn;

            int topRow = 0;
            if (this.menuBar.Visible)
                topRow = this.menuBar.Height;
            if (this.fileComboBar.Visible)
            {
                this.fileComboBar.Location = new Point(this.fileComboBar.Location.X, topRow);
                topRow += this.fileComboBar.Height;
            }
            if (this.filePathBar.Visible)
            {
                this.filePathBar.Location = new Point(this.filePathBar.Location.X, topRow);
                topRow += this.filePathBar.Height;
            }
            this.fileView.Location = new Point(this.fileView.Location.X, topRow);
            this.fileView.Height = this.statusBar.Location.Y - topRow;
        }

        /// <summary>
        /// Calculate column width based on previous percent layout.
        /// </summary>
        private void FileListExplorer_Resize(object sender, EventArgs e)
        {
            if (this.fileView.Columns.Count != 0)
            {
                int colCnt   = this.fileView.Columns.Count;
                int colWidth = this.fileView.Width / colCnt;
                ListView.ColumnHeaderCollection cols = this.fileView.Columns;
               
                foreach (ColumnHeader header in cols) 
                {
                    if (this.prevWidth == 0)
                        header.Width = colWidth;
                    else
                        header.Width = (int)((float)this.fileView.Width * header.Width / this.prevWidth);
                }

                this.prevWidth = this.fileView.Width;
            }
        }

        /// <summary>
        /// View change requested.
        /// </summary>
        private void ViewBtn_Click(object sender, EventArgs e)
        {
            ToolStripDropDownItem viewItem = (ToolStripDropDownItem)sender;
            ToolStripDropDownMenu viewItems = viewItem.DropDown.OwnerItem.Owner as ToolStripDropDownMenu;

            foreach (ToolStripDropDownItem item in viewItems.Items)
                item.Image = null;

            viewItem.Image = global::AGifCreator.Properties.Resources.greenCheck32;
            viewItems.OwnerItem.Text = viewItem.Text + "..."; ;
            this.fileView.View = (View)viewItem.Tag;
        }

        /// <summary>
        /// Column clicked - fire sort.
        /// </summary>
        private void fileView_ColumnClick(object sender, ColumnClickEventArgs e)
        {
            // Determine if clicked column is already the column that is being sorted.
            if (e.Column == this.columnSorter.SortColumn)
            {
                // Reverse the current sort direction for this column.
                if (columnSorter.Order == SortOrder.Ascending)
                    columnSorter.Order = SortOrder.Descending;
                else
                    columnSorter.Order = SortOrder.Ascending;
            }
            else
            {
                // Set the column number that is to be sorted; default to ascending.
                columnSorter.SortColumn = e.Column;
                columnSorter.Order = SortOrder.Ascending;
            }

            // Perform the sort with these new sort options.
            this.fileView.Sort();
        }

        /// <summary>
        /// Combo changed - update other components.
        /// </summary>
        private void fileComboBar_Changed(object sender, EventArgs e)
        {
            try
            {
                if (this.fileComboBar.FileName.Length > 0)
                {
                    DirectoryInfo dirInfo = new DirectoryInfo(this.fileComboBar.FileName);
                    this.Add(dirInfo);
                }
            }
            catch { }
        }

        /// <summary>
        /// Path Bar changed - update other components.
        /// </summary>
        private void filePathBar_Changed(object sender, EventArgs e)
        {
            try
            {
                // this.fileComboBar.FileName = this.filePathBar.FilePath;
                DirectoryInfo dirInfo = new DirectoryInfo(this.filePathBar.FilePath);
                this.Add(dirInfo);
            }
            catch { }
        }


        private int prevSender = -1;
        private long prevTick = DateTime.Now.Ticks;
        private void fileView_Click(object sender, EventArgs e)
        {
            int idx = this.fileView.SelectedIndices[0];

            long delta = DateTime.Now.AddTicks(-prevTick).TimeOfDay.Seconds;
            if (prevSender == idx && delta < 2)
                fileView_DoubleClick(sender, e);    // emulate Double click since it fails to function

            
            prevTick = DateTime.Now.Ticks;
            prevSender = idx;
        }

        /// <summary>
        /// Double click - if on directory open it, else fire change dbl click event.
        /// </summary>
        private void fileView_DoubleClick(object sender, EventArgs e)
        {
            System.Diagnostics.Debug.WriteLine("** Double **  Click");
            if (this.fileView.SelectedItems.Count == 1)
            {
                try
                {
                    DirectoryInfo dirInfo = new DirectoryInfo(this.fileView.SelectedItems[0].Name);
                    if (dirInfo.Exists)
                    {
                        Add(dirInfo);
                        return;
                    }
                }
                catch { }
            }

            if (this.DoubleClicked != null)
                this.DoubleClicked(this, EventArgs.Empty);
        }

        /// <summary>
        /// Process key commands (space = add)
        /// </summary>
        private void fileView_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Space:
                    if (this.DoubleClicked != null)
                        this.DoubleClicked(this, EventArgs.Empty);
                    break;
            }
        }
        #endregion

        #region Interface to update File List
        /// <summary>
        /// Append file to current view.
        /// </summary>
        const int maxNumImages = 400;
        const int largeImgDim = 48;
        public void Append(FileInfo file)
        {
            int imageIdx = smallImageList.Images.Count;

            if (imageIdx < maxNumImages)
            {
                bool needImage = true;
                const string validExt = ".bmp,.gif,.jpg,.jpeg,.png,.tif,.tiff,.wmf";

                if (validExt.Contains(file.Extension))
                {
                    try
                    {
                        Image image = Bitmap.FromFile(file.FullName);
                        largeImageList.Images.Add(image);
                        smallImageList.Images.Add(image);
                        needImage = false;
                    }
                    catch { }
                }

                if (needImage)
                {
                    try
                    {
                        // Extract an icon
                        Icon fileIcon = System.Drawing.Icon.ExtractAssociatedIcon(file.FullName);
                        largeImageList.Images.Add(fileIcon);
                        smallImageList.Images.Add(fileIcon);
                        needImage = false;
                    } catch {}
                }

                if (needImage)
                {
                    Icon fileIcon = global::AGifCreator.Properties.Resources.document;
                    largeImageList.Images.Add(fileIcon);
                    smallImageList.Images.Add(fileIcon);
                }
            }
            else if (imageIdx == maxNumImages)
            {
                Icon fileIcon = global::AGifCreator.Properties.Resources.document;
                largeImageList.Images.Add(fileIcon);
                smallImageList.Images.Add(fileIcon);
            }
            else
                imageIdx = maxNumImages;

            ListViewItem lvi =
                this.fileView.Items.Add(file.Name, file.Name, imageIdx);
            lvi.SubItems.Add(file.Length.ToString());
            lvi.SubItems.Add(file.Extension);
            lvi.SubItems.Add(file.LastWriteTime.ToShortDateString() + " " +
                file.LastWriteTime.ToShortTimeString());
            lvi.SubItems.Add(file.Attributes.ToString());
            lvi.Name = file.FullName;
        }

        /// <summary>
        /// Append directories to current  view.
        /// </summary>
        /// <param name="dir"></param>
        public void Append(DirectoryInfo dir)
        {
            int imageIdx = smallImageList.Images.Count;
            Icon fileIcon = global::AGifCreator.Properties.Resources.folder;
            smallImageList.Images.Add(fileIcon);
            largeImageList.Images.Add(fileIcon);

            ListViewItem lvi =
                this.fileView.Items.Add(dir.Name, dir.Name, imageIdx);
            lvi.SubItems.Add("");
            lvi.SubItems.Add("dir");
            lvi.SubItems.Add(dir.LastWriteTime.ToShortDateString() + " " +
                dir.LastWriteTime.ToShortTimeString());
            lvi.SubItems.Add(dir.Attributes.ToString());
            lvi.Name = dir.FullName;
        }

        /// <summary>
        /// Add directory and its files to  view.
        /// </summary>
        public void Add(DirectoryInfo dirInfo)
        {
            this.filePathBar.FilePath = dirInfo.FullName;
            this.fileComboBar.FileName = this.filePathBar.FilePath;

            // Fill file view
            this.fileView.BeginUpdate();
            this.fileView.Items.Clear();
            this.fileView.ListViewItemSorter = null;
            this.fileView.SmallImageList = null;
            this.fileView.LargeImageList = null;

            smallImageList = new ImageList();
            largeImageList = new ImageList();
            largeImageList.ImageSize = new Size(largeImgDim, largeImgDim);

            try
            {
                foreach (DirectoryInfo dir in dirInfo.GetDirectories())
                {
                    Append(dir);
                }
            }
            catch { }

            int dirCnt = this.fileView.Items.Count;

            try
            {
                foreach (FileInfo file in dirInfo.GetFiles())
                {
                    Append(file);
                }
            }
            catch { }

            this.columnSorter.Order = SortOrder.None;
            this.fileView.ListViewItemSorter = this.columnSorter;
            this.fileView.SmallImageList = this.smallImageList;
            this.fileView.LargeImageList = this.largeImageList;

            this.fileView.EndUpdate();

            int fileCnt = this.fileView.Items.Count - dirCnt;
            this.statusBar.Text = fileCnt.ToString() + " files";
            if (dirCnt != 0)
                this.statusBar.Text += ", " + dirCnt.ToString() + " directories";

            if (this.Changed != null)
                this.Changed(this, EventArgs.Empty);
        }

        /// <summary>
        /// Clear ui compents
        /// </summary>
        public void Clear()
        {
            this.fileView.Items.Clear();
            this.filePathBar.Text = string.Empty;
            this.statusBar.Text = string.Empty;
        }
        #endregion

        #region Read access to file list
        /// <summary>
        /// Return array of selected files.
        /// </summary>
        public FileInfo[] SelectedItems
        {
            get 
            {
                FileInfo[] fileInfo = new FileInfo[this.fileView.SelectedItems.Count];
                int idx = 0;
                foreach (ListViewItem item in this.fileView.SelectedItems)
                    fileInfo[idx++] = new FileInfo(item.Name as string);
                return fileInfo;
            }
        }

        /// <summary>
        /// Return array of files
        /// </summary>
        public FileInfo[] Items
        {
            get
            {
                FileInfo[] fileInfo = new FileInfo[this.fileView.Items.Count];
                int idx = 0;
                foreach (ListViewItem item in this.fileView.Items)
                    fileInfo[idx++] = new FileInfo(item.Name as string);
                return fileInfo;
            }
        }

        /// <summary>
        /// Return file path
        /// </summary>
        public DirectoryInfo Directory
        {
            get { return new DirectoryInfo(this.filePathBar.FilePath); }
        }
        #endregion

        #region Data Members
        private ImageList smallImageList;
        private ImageList largeImageList;
        private int prevWidth = 0;
        public event EventHandler Changed;
        public event EventHandler DoubleClicked;
        private ListViewColumnSorter columnSorter;
        #endregion

        #region Drag and Drop
        /// <summary>
        /// Forward drop to combobox
        /// </summary>
        private void FileListExplorer_DragDrop(object sender, DragEventArgs e)
        {
            this.fileComboBar.comboBox_DragDrop(sender, e);
        }

        /// <summary>
        /// Forward drag enter to combobox
        /// </summary>
        private void FileListExplorer_DragEnter(object sender, DragEventArgs e)
        {
            this.fileComboBar.comboBox_DragEnter(sender, e);
        }

        /// <summary>
        /// Start drag
        /// </summary>
        private void fileView_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Clicks != 0)
                System.Diagnostics.Debug.WriteLine("Mouse Move click=" + e.Clicks.ToString());

            if (e.Button == MouseButtons.Left /* && e.Clicks == 1 */)
            {
                if (this.fileView.SelectedItems.Count != 0)
                {
                    // Drag selected items.

                    string[] fileNames = new string[this.fileView.SelectedItems.Count];
                    int fn = 0;
                    foreach (ListViewItem lvi in this.fileView.SelectedItems)
                    {
                        fileNames[fn++] = lvi.Name;
                    }

                    DataObject dataObject = new DataObject(DataFormats.FileDrop, fileNames);
                    if (DoDragDrop(dataObject, DragDropEffects.Copy | DragDropEffects.Move) ==
                        DragDropEffects.Move)
                    {
                        // this.fileView.Items.Remove(item);
                    }
                }
                else
                {
                    ListViewItem item =
                        this.fileView.GetItemAt(e.X, e.Y);

                    if (item != null)
                    {
                        if (DoDragDrop(item.Name, DragDropEffects.Copy | DragDropEffects.Move) ==
                            DragDropEffects.Move)
                        {
                            // this.fileView.Items.Remove(item);
                        }
                    }
                }
            }
        }
        #endregion

    }
}
