﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace AGifCreator
{
    /// <summary>
    /// Manage list of images as a sequence of frames.
    ///   
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// </summary>
    public partial class ImageListExplorer : UserControl
    {
        public ImageListExplorer()
        {
            InitializeComponent();
            InitImageList(48, 128);
            this.sizeMenuBtn.Text = "Size 128 x 128";
            this.frameMenuBtn.Text = "Frames " + this.maxFrames.ToString();

            ViewBtn_Click(this.thumbViewBtn, EventArgs.Empty);
            ViewStyle(true, false, false);

            // Create an instance of a ListView column sorter and assign it 
            // to the ListView control.
            this.columnSorter = new ListViewColumnSorter();
            this.fileView.ListViewItemSorter = this.columnSorter;

            ListViewHelper.EnableDoubleBuffer(this.fileView);
        }

        #region User Interface callbacks
        /// <summary>
        /// Adjust view list style (detail, list, icons, thumbs)
        /// </summary>
        public void ViewStyle(bool menuBarOn, bool comboOn, bool pathBarOn)
        {
            this.menuBar.Visible = menuBarOn;
            this.fileComboBar.Visible = comboOn;
            this.filePathBar.Visible = pathBarOn;

            int topRow = 0;
            if (this.menuBar.Visible)
                topRow = this.menuBar.Height;
            if (this.fileComboBar.Visible)
            {
                this.fileComboBar.Location = new Point(this.fileComboBar.Location.X, topRow);
                topRow += this.fileComboBar.Height;
            }
            if (this.filePathBar.Visible)
            {
                this.filePathBar.Location = new Point(this.filePathBar.Location.X, topRow);
                topRow += this.filePathBar.Height;
            }
            this.fileView.Location = new Point(this.fileView.Location.X, topRow);
            this.fileView.Height = this.statusBar.Location.Y - topRow;
        }

        /// <summary>
        /// Update column widths to keep relative layout constant.
        /// </summary>
        private void ImageListExplorer_Resize(object sender, EventArgs e)
        {
            if (this.fileView.Columns.Count != 0)
            {
                int colCnt = this.fileView.Columns.Count;
                int colWidth = this.fileView.Width / colCnt;
                ListView.ColumnHeaderCollection cols = this.fileView.Columns;

                foreach (ColumnHeader header in cols)
                {
                    if (this.prevWidth == 0)
                        header.Width = colWidth;
                    else
                        header.Width = (int)((float)this.fileView.Width * header.Width / this.prevWidth);
                }

                this.prevWidth = this.fileView.Width;
            }
        }

        private void ViewBtn_Click(object sender, EventArgs e)
        {
            ToolStripDropDownItem viewItem = (ToolStripDropDownItem)sender;
            ToolStripDropDownMenu viewItems = viewItem.DropDown.OwnerItem.Owner as ToolStripDropDownMenu;

            foreach (ToolStripDropDownItem item in viewItems.Items)
                item.Image = null;

            viewItem.Image = global::AGifCreator.Properties.Resources.greenCheck32;
            viewItems.OwnerItem.Text = viewItem.Text + "..."; ;
            this.fileView.View = (View)viewItem.Tag;
        }

        private void fileView_ColumnClick(object sender, ColumnClickEventArgs e)
        {
            // Determine if clicked column is already the column that is being sorted.
            if (e.Column == this.columnSorter.SortColumn)
            {
                // Reverse the current sort direction for this column.
                if (columnSorter.Order == SortOrder.Ascending)
                    columnSorter.Order = SortOrder.Descending;
                else
                    columnSorter.Order = SortOrder.Ascending;
            }
            else
            {
                // Set the column number that is to be sorted; default to ascending.
                columnSorter.SortColumn = e.Column;
                columnSorter.Order = SortOrder.Ascending;
            }

            // Perform the sort with these new sort options.
            this.fileView.Sort();
        }
  
        private void fileComboBar_Changed(object sender, EventArgs e)
        {
            try
            {
                DirectoryInfo dirInfo = new DirectoryInfo(this.fileComboBar.FileName);
                this.Add(dirInfo);
            }
            catch { }
        }


        private void fileView_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                this.popupMenu.Show(this.fileView, e.Location);
            }
        }

        private void fileView_DoubleClick(object sender, EventArgs e)
        {
            if (this.DoubleClicked != null)
                this.DoubleClicked(this, EventArgs.Empty);
        }

        private void fileView_ItemMouseHover(object sender, ListViewItemMouseHoverEventArgs e)
        {
            this.ViewNotify(e.Item);
        }

        private void fileView_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (this.fileView.SelectedItems.Count > 0)
                ViewNotify(this.fileView.SelectedItems[0]);
        }

        private void ViewNotify(ListViewItem item)
        {
            foreach (ListViewItem lvi in this.fileView.Items)
                lvi.BackColor = Color.White;

            // item.Selected = true;
            item.EnsureVisible();
            item.BackColor = Color.Coral;

            this.statusBar.Text = "Frame:" +
                    item.SubItems[0].Text + ", " +   // frame
                    item.SubItems[1].Text + "  (" +  // file
                    item.SubItems[3].Text + " x " +  // width
                    item.SubItems[4].Text + ")";     // hiehgt

            if (this.HoverChanged != null)
                this.HoverChanged(this, new ListViewItemMouseHoverEventArgs(item));
        }
        #endregion

        /// <summary>
        /// Initialize image lists, add default images.
        /// </summary>
        private void InitImageList(int smallSz, int largeSz)
        {
            if (smallImageList == null)
            {
                smallImageList = new ImageList();
                largeImageList = new ImageList();

                smallImageList.ImageSize = new Size(smallSz, smallSz);
                largeImageList.ImageSize = new Size(largeSz, largeSz);

                // Add folder as default first image
                Icon fileIcon = global::AGifCreator.Properties.Resources.folder;
                smallImageList.Images.Add(fileIcon);
                largeImageList.Images.Add(fileIcon);

                // Add sand timer as defafult 2nd image.
                Image sandTimerS = global::AGifCreator.Properties.Resources.sand_timerS;
                Image sandTimerA = global::AGifCreator.Properties.Resources.sand_timerA;
                smallImageList.Images.Add(sandTimerS);
                largeImageList.Images.Add(sandTimerA);

                this.fileView.SmallImageList = smallImageList;
                this.fileView.LargeImageList = largeImageList;
            }
        }

        #region Interface to update image list.


        /// <summary>
        /// Append image to frame list. If image is animate GIF, append each frame.
        /// </summary>
        public bool Append(FileInfo fileInfo)
        {
            try
            {
                if (smallImageList.Images.Count >= 100)
                    return Append(fileInfo, null, -1);

                // Load image
                Image image = Bitmap.FromFile(fileInfo.FullName);
                image.Tag = fileInfo.FullName;

                int frameCount = 0;
                try
                {
                    FrameDimension oDimension = new FrameDimension(image.FrameDimensionsList[0]);
                    frameCount = image.GetFrameCount(oDimension);

                    if (frameCount > 1)
                    {
                        // if image contains multiple frames - extract each frame.
                        for (int frameIdx = 0; frameIdx < frameCount; frameIdx++)
                        {
                            image.SelectActiveFrame(oDimension, frameIdx);

                            Image imageFrame = new Bitmap(image.Width, image.Height, image.PixelFormat);
                            imageFrame.Tag = fileInfo.FullName;
                            Graphics g = Graphics.FromImage(imageFrame);
                            g.DrawImageUnscaled(image, new Point(0, 0));
                            Append(fileInfo, imageFrame, frameIdx);
                        }

                        this.sPlayer.Stream = Properties.Resources.click1;
                        this.sPlayer.Play();

                        return true;
                    }
                }
                catch { };

                Append(fileInfo, image, -1);

                this.sPlayer.Stream = Properties.Resources.click1;
                this.sPlayer.Play();

                return true;
            }
            catch 
            {
                this.statusBar.Text = "Ignoring Image " + fileInfo.Name;

                this.sPlayer.Stream = Properties.Resources.drum1;
                this.sPlayer.Play();
            }

            return false;
        }

        public bool Append(FileInfo file, Image image, int userId)
        {
            try
            {
                // Extract an image
                int imageIdx = (image == null) ? 1 : smallImageList.Images.Count;

                string frameName = this.fileView.Items.Count.ToString();

                ListViewItem lvi =
                    this.fileView.Items.Add(file.Name, frameName, imageIdx);
                lvi.Name = file.FullName;
                lvi.SubItems.Add(file.Name);
                lvi.SubItems.Add(file.Extension);

                if (imageIdx != 1)
                {
                    // Load image
                    // Image image = Bitmap.FromFile(file.FullName);

                    lvi.SubItems.Add(image.Width.ToString());
                    lvi.SubItems.Add(image.Height.ToString());
                    string userStr = "";
                    if (userId != -1)
                        userStr = userId.ToString() + " ";
                    lvi.SubItems.Add(userStr + image.PixelFormat.ToString());
                    lvi.SubItems.Add(file.LastWriteTime.ToShortDateString() + " " +
                        file.LastWriteTime.ToShortTimeString());
                    lvi.SubItems.Add(file.Attributes.ToString());

                    smallImageList.Images.Add(image);
                    largeImageList.Images.Add(image);
                }
                else
                {
                    // Skip loading image. - too many images.
                    lvi.SubItems.Add("width?");
                    lvi.SubItems.Add("height?");
                    lvi.SubItems.Add("pixel?");
                    lvi.SubItems.Add(file.LastWriteTime.ToShortDateString() + " " +
                        file.LastWriteTime.ToShortTimeString());
                    lvi.SubItems.Add(file.Attributes.ToString());
                }

                return true;
            }
            catch
            {
                this.statusBar.Text = "Ignoring Image " + file.Name;
            }

            return false;
        }

        public void Append(DirectoryInfo dir)
        {
            int imageIdx = 0;

            ListViewItem lvi =
                this.fileView.Items.Add(dir.Name, dir.Name, imageIdx);
            lvi.Name = dir.FullName;
            lvi.SubItems.Add(dir.Name);
            lvi.SubItems.Add("dir");
            lvi.SubItems.Add(dir.LastWriteTime.ToShortDateString() + " " +
                dir.LastWriteTime.ToShortTimeString());
            lvi.SubItems.Add(dir.Attributes.ToString());
        }

        public void Add(DirectoryInfo dirInfo)
        {
            this.filePathBar.Text = dirInfo.FullName;

            // Fill file view
            this.fileView.BeginUpdate();
            this.fileView.Items.Clear();

            smallImageList = null;

            foreach (DirectoryInfo dir in dirInfo.GetDirectories())
            {
                Append(dir);
            }
            int dirCnt = this.fileView.Items.Count;

            foreach (FileInfo file in dirInfo.GetFiles())
            {
                Append(file);
            }

            this.fileView.EndUpdate();

            int fileCnt = this.fileView.Items.Count - dirCnt;
            this.statusBar.Text = fileCnt.ToString() + " files";
            if (dirCnt != 0)
                this.statusBar.Text += ", " + dirCnt.ToString() + " directories";

            if (this.ListChanged != null)
                this.ListChanged(this, EventArgs.Empty);
        }

        public void Renumber()
        {
            int cnt = this.fileView.Items.Count;
            for (int idx = 0; idx < cnt; idx++)
            {
                this.fileView.Items[idx].Text = idx.ToString();
            }
        }
        public void ResizeImages(Size size)
        {
            if (this.largeImageList.ImageSize != size)
            {
                ImageList newImageList = new ImageList();
                newImageList.ImageSize = size;

                foreach (Image image in this.largeImageList.Images)
                {
                    newImageList.Images.Add(image);
                }

                this.largeImageList = newImageList;
                this.fileView.LargeImageList = newImageList;
                this.sizeMenuBtn.Text = "Size " + SizeForm.ToString(size);
            }
        }
        private void ImageZoom(float percent)
        {
            this.ResizeImages(new Size(
                 Math.Min(256, Math.Max(8, (int)(this.largeImageList.ImageSize.Width * percent))),
                 Math.Min(256, Math.Max(8, (int)(this.largeImageList.ImageSize.Height * percent)))));
        }


        /// <summary>
        /// Delete image from list by matching filename.
        /// </summary>
        public void Delete(FileInfo fileInfo)
        {
            ListViewItem[] matches = this.fileView.Items.Find(fileInfo.FullName, false);
            foreach (ListViewItem item in matches)
            {
                DeleteItem(item);
            }

            Renumber();
        }

        /// <summary>
        /// Delete all select images.
        /// </summary>
        public void DeleteSelected()
        {
            foreach (ListViewItem item in this.fileView.SelectedItems)
            {
                DeleteItem(item);
            }

            Renumber();
        }

        /// <summary>
        /// Delete image from list, removing images from large and small image lists.
        /// </summary>
        public void DeleteItem(ListViewItem item)
        {
            int imageIndex = item.ImageIndex;
            if (imageIndex > 1)
            {
                // remove any custom images
                this.largeImageList.Images[imageIndex].Dispose();
                this.smallImageList.Images[imageIndex].Dispose();
            }
            this.fileView.Items.Remove(item);
        }

        /// <summary>
        /// Remove all images from list, clear pathBar and statusBar.
        /// </summary>
        public void Clear()
        {
            this.fileView.Items.Clear();
            this.filePathBar.Text = string.Empty;
            this.statusBar.Text = string.Empty;
        }
        #endregion
 
        #region Read access to Image list.

        /// <summary>
        ///  Return access to ImageList.
        /// </summary>
        public ImageList ImageList
        {
            get 
            {
                // Copy images out in Frame order
                ImageList imageList = new ImageList();
                imageList.ImageSize = this.largeImageList.ImageSize;

                foreach (ListViewItem lvi in this.fileView.Items)
                    imageList.Images.Add(this.largeImageList.Images[lvi.ImageIndex]);

                return imageList;
            }
        }

        public FileInfo[] SelectedItems
        {
            get
            {
                FileInfo[] fileInfo = new FileInfo[this.fileView.SelectedItems.Count];
                int idx = 0;
                foreach (ListViewItem item in this.fileView.SelectedItems)
                    fileInfo[idx++] = new FileInfo(item.Name as string);
                return fileInfo;
            }
        }

        public FileInfo[] Items
        {
            get
            {
                FileInfo[] fileInfo = new FileInfo[this.fileView.Items.Count];
                int idx = 0;
                foreach (ListViewItem item in this.fileView.Items)
                    fileInfo[idx++] = new FileInfo(item.Name as string);
                return fileInfo;
            }
        }

        public int Count
        {
            get { return this.fileView.Items.Count; }
        }

        public int Frame
        {
            get { return (this.fileView.SelectedIndices.Count > 0) ?  this.fileView.SelectedIndices[0] : -1; }
            set { if ((UInt32)value < Count) this.ViewNotify(this.fileView.Items[value]); }
        }

        public DirectoryInfo Directory
        {
            get { return new DirectoryInfo(this.filePathBar.Text); }
        }

        public Image ImageFor(ListViewItem item)
        {
            return largeImageList.Images[item.ImageIndex];
        }

        #endregion

        #region Drag and Drop 
        /// <summary>
        /// Drag drop - process text or filedrops.
        /// </summary>
        private void fileView_DragDrop(object sender, DragEventArgs e)
        {
            string[] fileNames = null;
            if (e.Data.GetDataPresent(DataFormats.Text))
            {
                fileNames = new string[1];
                fileNames[0] = (string)e.Data.GetData(DataFormats.Text);
            }
            else if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                fileNames = (string[])e.Data.GetData(DataFormats.FileDrop);
            }

            Point ourPos = this.fileView.PointToClient(new Point(e.X, e.Y));
            ListViewItem dropItem =
                this.fileView.GetItemAt(ourPos.X, ourPos.Y);

            foreach (string fileName in fileNames)
            {
                if (fileName != null && fileName.Length > 0)
                {
                    try
                    {
                        FileInfo fileInfo = new FileInfo(fileName);
                        if (fileInfo.Exists)
                        {
                            // Always append  dropped  item.
                            this.Append(fileInfo);
                            int newItemIdx = this.fileView.Items.Count - 1;
                            if (dropItem != null && dropItem.Index + 1 < newItemIdx)
                            {
                                // Move dropped item to insertion point.
                                ListViewItem item = this.fileView.Items[newItemIdx];
                                this.fileView.Items.RemoveAt(newItemIdx);
                                this.fileView.Items.Insert(dropItem.Index, item);
                            }

                            Renumber();
                        }
                    }
                    catch { }
                }
            }
        }

        /// <summary>
        /// Drag enter - Support text and file drops.
        /// </summary>
        private void fileView_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.Text) ||
                e.Data.GetDataPresent(DataFormats.FileDrop) ||
                e.Data.GetDataPresent(DataFormats.StringFormat))
            {
                int keyShift = 4;
                e.Effect = ((e.KeyState & keyShift) != 0) ? DragDropEffects.Move : DragDropEffects.Copy;
            }
            else
            {
                e.Effect = DragDropEffects.None;
                System.Diagnostics.Debug.WriteLine(e.Data.ToString());
            }
        }

        /// <summary>
        /// Start a drag (mouse move with left button down)
        /// </summary>
        private void fileView_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left /* && e.Clicks == 1 */)
            {
                ListViewItem item =
                    this.fileView.GetItemAt(e.X, e.Y);

                if (item != null)
                {
                    if (DoDragDrop(item.Name, DragDropEffects.Copy | DragDropEffects.Move) ==
                        DragDropEffects.Move)
                    {
                        this.fileView.Items.Remove(item);
                        Renumber();
                    }
                }
            }
        }

        #endregion

        #region Prompt for maximum frame count, update frame list.
        /// <summary>
        /// Simple dialog to collect frame number
        /// </summary>
        public class FrameForm : Form
        {
            public FrameForm(string titleStr, string labelStr, int value)
            {
                textBox = new TextBox();

                Label lbl = new Label();
                Button okBtn = new Button();
                Button cancelBtn = new Button();

                this.Text = titleStr;
                textBox.Text = value.ToString();
                lbl.Text = labelStr;
                okBtn.Text = "Accept";
                cancelBtn.Text = "Cancel";

                this.Size = new Size(200, 100);
                lbl.Location = new Point(5, 5);
                textBox.Location = new Point(5+lbl.Width+5, 5);
                okBtn.Location = new Point(5, 30);
                cancelBtn.Location = new Point(100, 30);

                this.SuspendLayout();
                this.Controls.Add(lbl);
                this.Controls.Add(textBox);
                this.Controls.Add(okBtn);
                this.Controls.Add(cancelBtn);
                this.ResumeLayout();

                okBtn.Click += new EventHandler(okBtn_Click);
                cancelBtn.Click += new EventHandler(cancelBtn_Click);
                this.MouseMove += new System.Windows.Forms.MouseEventHandler(this.mouseMove);
            }

            public TextBox textBox;

            public void okBtn_Click(object sender, EventArgs e)
            {
                this.DialogResult = DialogResult.OK;
                Close();
            }
            public void cancelBtn_Click(object sender, EventArgs e)
            {
                this.DialogResult = DialogResult.Cancel;
                Close();
            }

            private Point mouseStartAt = Point.Empty;
            private void mouseMove(object sender, MouseEventArgs e)
            {
                if (e.Button == MouseButtons.Left)
                {
                    if (mouseStartAt == Point.Empty)
                    {
                        mouseStartAt = e.Location;
                        this.textBox.BackColor = Color.Aqua;
                    }
                    else
                    {
                        int off = Math.Abs(e.X - mouseStartAt.X) > Math.Abs(e.Y - mouseStartAt.Y) ?
                            (e.X - mouseStartAt.X) : (e.Y - mouseStartAt.Y);
                        this.textBox.Text = Math.Max(0, Int32.Parse(textBox.Text) + off/10).ToString();
                    }
                }
                else
                {
                    mouseStartAt = Point.Empty;
                    this.textBox.BackColor = Color.White;
                }
            }
        }

        /// <summary>
        /// Add points together, return sum.
        /// </summary>
        private Point Add(Point p1, Point p2)
        {
            return new Point(p1.X + p2.X, p1.Y + p2.Y);
        }

        /// <summary>
        /// Prompt for new maximum frame count.
        /// </summary>
        private void frameMenuBtn_Click(object sender, EventArgs e)
        {
            FrameForm frameForm = new FrameForm("Max Frames", "Frames:", this.maxFrames);
            frameForm.Location = Add(this.PointToScreen(this.Location), new Point(50,50));
            frameForm.StartPosition = System.Windows.Forms.FormStartPosition.Manual;

            if (frameForm.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    this.maxFrames = Int32.Parse(frameForm.textBox.Text);
                    int cnt = fileView.Items.Count;
                    while (cnt > this.maxFrames)
                    {
                        this.DeleteItem(this.fileView.Items[--cnt]);
                    }

                    this.frameMenuBtn.Text = "Frames " + this.maxFrames.ToString();
                }
                catch { }
            }
        }
        #endregion


        #region Size Dialog 
        /// <summary>
        /// Dailog box to get image size.
        /// </summary>
        public class SizeForm : Form
        {
            public SizeForm(string titleStr, string labelStr, Size value)
            {
                textBox = new TextBox();

                Label lbl = new Label();
                Button okBtn = new Button();
                Button cancelBtn = new Button();

                this.Text = titleStr;
                textBox.Text = ToString(value);
                lbl.Text = labelStr;
                okBtn.Text = "Accept";
                cancelBtn.Text = "Cancel";

                this.Size = new Size(200, 100);
                lbl.Location = new Point(5, 5);
                textBox.Location = new Point(5 + lbl.Width + 5, 5);
                okBtn.Location = new Point(5, 30);
                cancelBtn.Location = new Point(100, 30);

                this.SuspendLayout();
                this.Controls.Add(lbl);
                this.Controls.Add(textBox);
                this.Controls.Add(okBtn);
                this.Controls.Add(cancelBtn);
                this.ResumeLayout();

                okBtn.Click += new EventHandler(okBtn_Click);
                cancelBtn.Click += new EventHandler(cancelBtn_Click);
                this.MouseMove += new System.Windows.Forms.MouseEventHandler(this.mouseMove);
            }

            public TextBox textBox;

            public static string ToString(Size s)
            {
                return s.Width.ToString() + " x " + s.Height.ToString();
            }

            public void okBtn_Click(object sender, EventArgs e)
            {
                this.DialogResult = DialogResult.OK;
                Close();
            }
            public void cancelBtn_Click(object sender, EventArgs e)
            {
                this.DialogResult = DialogResult.Cancel;
                Close();
            }

            public Size GetSize()
            {
                string[] fields = textBox.Text.Split('x');
                return new Size(Int32.Parse(fields[0]), Int32.Parse(fields[1]));
            }

            private Point mouseStartAt = Point.Empty;
            private void mouseMove(object sender, MouseEventArgs e)
            {
                if (e.Button == MouseButtons.Left)
                {
                    if (mouseStartAt == Point.Empty)
                    {
                        mouseStartAt = e.Location;
                        this.textBox.BackColor = Color.Aqua;
                    }
                    else
                    {
                        Size sz = GetSize();
                        sz = new Size(Math.Max(0, sz.Width + (e.X - mouseStartAt.X) / 10),
                            Math.Max(0, sz.Height + (e.Y - mouseStartAt.Y) / 10));
                        this.textBox.Text = ToString(sz);
                    }
                }
                else
                {
                    mouseStartAt = Point.Empty;
                    this.textBox.BackColor = Color.White;
                }
            }
        }
 
        /// <summary>
        /// Show size dialog - collect new size, update image buffers.
        /// </summary>
        /// 
        private void xdoubleSizeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ImageZoom(2.0f);
        }

        private void reduceSizeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ImageZoom(0.5f);
        }

        private void customSizeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SizeForm sizeForm = new SizeForm("Image Size", "Size(w x h):",
                this.largeImageList.ImageSize);
            sizeForm.Location = Add(this.PointToScreen(this.Location), new Point(50, 50));
            sizeForm.StartPosition = System.Windows.Forms.FormStartPosition.Manual;

            if (sizeForm.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    string[] fields = sizeForm.textBox.Text.Split('x');
                    int w = Int32.Parse(fields[0]);
                    int h = Int32.Parse(fields[1]);
                    this.ResizeImages(new Size(w, h));
                    Size sz = this.largeImageList.ImageSize;
                    this.sizeMenuBtn.Text = "Size " + SizeForm.ToString(sz);
                }
                catch { }
            }
        }

        #endregion

        #region Data members

        private ImageList smallImageList;
        private ImageList largeImageList;
        private int prevWidth = 0;
        public event EventHandler HoverChanged;
        public event EventHandler ListChanged;
        public event EventHandler DoubleClicked;
        private ListViewColumnSorter columnSorter;
        private int maxFrames = 100;
        private System.Media.SoundPlayer sPlayer = new System.Media.SoundPlayer();

        #endregion

        private void SwapFrame(int frameIdx1, int frameIdx2)
        {
            if (frameIdx1 >= 0 && frameIdx2 < this.fileView.Items.Count)
            {
                ListViewItem item1 = this.fileView.Items[frameIdx1];
                ListViewItem item2 = this.fileView.Items[frameIdx2];
                this.fileView.Items.Remove(item2);
                this.fileView.Items.Remove(item1);
                this.fileView.Items.Insert(frameIdx1, item2);
                this.fileView.Items.Insert(frameIdx2, item1);

                Renumber();
            }
        }

        private void fileView_KeyDown(object sender, KeyEventArgs e)
        {
            bool alt = (e.Modifiers & Keys.Alt) != 0;
            bool ctl = (e.Modifiers & Keys.Control) != 0;
            int selectIdx = (this.fileView.SelectedIndices.Count > 0) ?
                this.fileView.SelectedIndices[0] : -1;

            switch (e.KeyCode)
            {
                case Keys.X:
                    this.DeleteSelected();
                    break;
                case Keys.Delete:
                    this.DeleteSelected();
                    break;
            }

            if (ctl)
            {
                switch (e.KeyCode)
                {
                    case Keys.Oemplus:
                        this.ImageZoom(2.0f);
                        break;
                    case Keys.OemMinus:
                        this.ImageZoom(0.5f);
                        break;
                }
            }
            if (alt)
            {
                switch (e.KeyCode)
                {
                    case Keys.Left:
                        this.SwapFrame(selectIdx - 1, selectIdx);
                        break;
                    case Keys.Right:
                        this.SwapFrame(selectIdx, selectIdx + 1);
                        break;
                    case Keys.Up:
                        this.SwapFrame(selectIdx - 1, selectIdx);
                        break;
                    case Keys.Down:
                        this.SwapFrame(selectIdx, selectIdx + 1);
                        break;
                }
            }
        }

        /// <summary>
        /// Popup menu to save frame image to disk.
        /// </summary>
        private void popupSaveAsBtn_Click(object sender, EventArgs e)
        {
            Point ourPos = this.fileView.PointToClient(Control.MousePosition);
            ListViewItem item =
                this.fileView.GetItemAt(ourPos.X, ourPos.Y);

            if (item != null && this.saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                Image image = this.largeImageList.Images[item.ImageIndex];
                image.Save(this.saveFileDialog.FileName);
                this.statusBar.Text = "Saved Frame " + item.Index.ToString() +
                    " As:" + this.saveFileDialog.FileName;
            }
        }
    }
}
