﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Collections;
using System.Diagnostics;

namespace AGifCreator
{
    /// <summary>
    /// Extract colors from True color image.
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public unsafe class ColorManager
    {
        private struct ARGBFreq
        {
            public Int32 count;
            public Int32 color;
        }

        private struct ColorFreq
        {
            public Int32 count;
            public Color color;
        }

        private class SortARGBFreq : IComparer
        {
            int IComparer.Compare(Object x, Object y)
            {
                // Sort reverse order so highest freq is first.
                ARGBFreq cf1 = (ARGBFreq)x;
                ARGBFreq cf2 = (ARGBFreq)y;
                return (cf2.count - cf1.count);
            }
        }

        private class SortColorBrightness : IComparer
        {
            int IComparer.Compare(Object x, Object y)
            {
                Color c1 = (Color)x;
                Color c2 = (Color)y;
                return c1.GetBrightness().CompareTo(c2.GetBrightness());
            }
        }

        private class SortColorRGB : IComparer
        {
            int IComparer.Compare(Object x, Object y)
            {
                Color c1 = (Color)x;
                Color c2 = (Color)y;
                uint v1 = RgbValue(c1);
                uint v2 = RgbValue(c2);
                return (v1 < v2) ? -1 : (v1 == v2) ? 0 : 1;
            }

            private uint RgbValue(Color c)
            {
                if (Math.Abs((int)c.R - (int)c.G) < 4 &&
                    Math.Abs((int)c.R - (int)c.B) < 4)
                {
                    // gray
                    return c.R;
                }
                else
                    if (c.R >= c.G && c.R >= c.B)
                        return c.R * 256u * 256u * 256u + c.G * 256u + c.B;
                    else if (c.G >= c.R && c.G >= c.B)
                        return c.G * 256u * 256u + c.B;
                    else
                        return c.B * 256u;
            }
        }

        /// <summary>
        /// Sort array of colors by RGB distance
        /// </summary>
        private Color[] SortColors(Color[] colors)
        {
            ArrayList cList = new ArrayList(colors);
            cList.Sort(new SortColorRGB());
            Color[] sorted = new Color[colors.Length];
            int n = 0;
            foreach (Color c in cList)
                sorted[n++] = c;
            return sorted;
        }

        public ArrayList GetColorFreq(Bitmap inImage)
        {
            ArrayList sortFreq = new ArrayList();

            // Get the source image bits and lock into memory
            int height = inImage.Height;
            int width = inImage.Width;
            Rectangle bounds = new Rectangle(0, 0, width, height);
            BitmapData sourceData = null;

            try
            {
                sourceData = inImage.LockBits(bounds, ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                byte* pSourceRow = (byte*)sourceData.Scan0.ToPointer();
                Int32* pSourcePixel;
                Hashtable colorMap = new Hashtable();
                Int32 zero = 0;
                Int32 maskARGB = 0x00ffffff; // remove alpha and lower 2 bits of each channel.

                // Count colors present in image.
                for (int row = 0; row < height; row++)
                {
                    pSourcePixel = (Int32*)pSourceRow;

                    // Skip columns to improve scan performance. 
                    int colInc = 1;
                    colInc = ((colInc + 2) % 11) + 1;    // 4, 7, 10, 2, 5, 8, 11, 3 ... 

                    for (int col = 0; col < width; col += colInc, pSourcePixel += colInc)
                    {
                        Int32 c = *pSourcePixel & maskARGB;
                        if (colorMap.Contains(c))
                            colorMap[c] = (Int32)colorMap[c] + 1;
                        else
                        {
                            colorMap.Add(c, zero);
                            if (colorMap.Count == 4096)
                            {
                                maskARGB = 0x00fcfcfc;
                            }
                            else if (colorMap.Count == 1024)
                            {
                                maskARGB = 0x00fefefe;
                            }
                        }
                    }

                    pSourceRow += sourceData.Stride;
                }

                // Transfer Hash to Array so we can sort by count.
                int nColors = colorMap.Keys.Count;
                foreach (Int32 n in colorMap.Keys)
                {
                    ARGBFreq argbFreq;
                    argbFreq.color = (Int32)n;
                    argbFreq.count = (Int32)colorMap[n];
                    if (argbFreq.count > 0)
                        sortFreq.Add(argbFreq);
                }

                sortFreq.Sort(new SortARGBFreq());
            }
            finally
            {
                // Ensure that the bits are unlocked
                inImage.UnlockBits(sourceData);
            }

            return sortFreq;
        }

        /// <summary>
        ///  Reduce sorted ARGB color frequence array by removing similar colors.
        /// </summary>
        /// <param name="sortArgbFreq">Array of ArgbFreq</param>
        /// <returns>Array of ColorFreq</returns>
        public ArrayList ReduceColors(ArrayList sortArgbFreq, int minDist)
        {
            ArrayList reducedFreq = new ArrayList();
            Int32 srtCnt = sortArgbFreq.Count;

            for (int srtIdx = 1; srtIdx < srtCnt && reducedFreq.Count < 256; srtIdx++)
            {
                ARGBFreq argbFreg = (ARGBFreq)sortArgbFreq[srtIdx];
                Int32 cARGB = argbFreg.color;
                int r = (byte)(cARGB >> 16);
                int g = (byte)(cARGB >> 8);
                int b = (byte)(cARGB);

                Int32 d = 0;
                for (int ic = 0; ic < reducedFreq.Count; ic++)
                {
                    Color cs = ((ColorFreq)reducedFreq[ic]).color;
                    d = Math.Abs(r - cs.R) + Math.Abs(g - cs.G) + Math.Abs(b - cs.B);
                    if (d < minDist)
                        break;
                }

                if (d > minDist)
                {
                    ColorFreq colorFreq = new ColorFreq();
                    colorFreq.color = Color.FromArgb(cARGB);
                    colorFreq.count = argbFreg.count;
                    reducedFreq.Add(colorFreq);
                }
            }

            return reducedFreq;
        }

        /// <summary>
        /// Extract a reasonable color Palette from a Truecolor image
        /// </summary>
        /// <param name="inImage">Input truecolor image</param>
        /// <returns>color array extracted from image</returns>
        public Color[] Palette(Bitmap inImage)
        {
            ColorPalette colorPalette = inImage.Palette;
            Color[] colors = colorPalette.Entries;

            if (colors.Length == 0)
            {
                // Okay - we failed to get a quick color palette, time to make one
                //   Get array of sorted colors sorted by frequence;
                //   Reduce array by removing similar colors
                //   Save best 256 colors.
                //   
                ArrayList sortFreq = GetColorFreq(inImage);
                ArrayList colorFreq = ReduceColors(sortFreq, 16);
                colors = new Color[256];
                int colorCnt = Math.Min(colors.Length, colorFreq.Count);

                for (int colorIdx = 0; colorIdx < colorCnt; colorIdx++)
                    colors[colorIdx] = ((ColorFreq)sortFreq[colorIdx]).color;
            }

            return colors;
        }
    }
}
